#!/usr/bin/env python
"""
This program is used to visualize contour data created by SimVascular. 
"""
import argparse
import logging
import sys
import xml.etree.ElementTree as et

import vtk
import numpy as np
import copy

class ContourGroup(object):
    """
    The ContourGroup class is used to represent a SimVascular contour group.
    
    Obs.: It is ignoring the timestep tag.
    """
    
    def __init__(self, logger):
        self.logger = logger
        self.attrib = dict()
        self.lofting_parameters = dict()
        self.contours = []
        
    def __repr__(self):
        return '<ContourGroup(%s) id=%s>' % (self.attrib["path_name"], id(self))
    
    def __copy__(self):
        return ContourGroup(self, self.logger)
    
    def __deepcopy__(self, memo):
        newContourGroup = ContourGroup(self.logger)
        newContourGroup.__dict__.update(self.__dict__)
        newContourGroup.attrib = copy.deepcopy(self.attrib, memo)
        newContourGroup.lofting_parameters = copy.deepcopy(self.lofting_parameters, memo)
        newContourGroup.contours = copy.deepcopy(self.contours, memo)
        
        return newContourGroup
    
    def read(self, file_name):
        """
        Read a ctgr file. 
        """
        
        if file_name == None:
            self.logger.error("No input contour file name given.")
            sys.exit(1)
        else:
            self.logger.info("Input contour file name: %s" % file_name)
    
        # Remove 'format' tag from xml file.
        f = open(file_name, "r")
        lines = f.readlines()
        new_lines = []
        for line in lines:
          if '<format' not in line:
            new_lines.append(line)
        # Create string from xml file and parse it.
        xml_string = "".join(new_lines)
        
        tree = et.ElementTree(et.fromstring(xml_string))
        contour_group_et = tree.getroot() #the root is contour_group
    
        self.attrib["path_name"] = contour_group_et.attrib.get("path_name")
        self.logger.info("Contour group: %s" % self.attrib["path_name"])
        
        self.attrib["path_id"] = contour_group_et.attrib.get("path_id")
        self.attrib["reslice_size"] = contour_group_et.get("reslice_size")
        self.attrib["point_2D_display_size"] = contour_group_et.attrib.get("point_2D_display_size")
        self.attrib["point_size"] = contour_group_et.attrib.get("point_size")
        
        lofting_parameters_et = contour_group_et.find("timestep/lofting_parameters")
        for attrib, value in lofting_parameters_et.attrib.items():
            self.lofting_parameters[attrib] = value
        
        for contour_et in contour_group_et.iter('contour'):
            contour = Contour()
            #logger.info("   Contour id: %s" % id)
            contour.read(contour_et)
            #_for control_pts in contour_t
            self.contours.append(contour)
        #_for contour_t in contour_group_t
        self.logger.info("   Number of contours %d" % len(self.contours))
    
    def write(self, file_name):
        """
        Write a ctgr file. 
        """
        
        contour_group_et = et.Element("contourgroup")
        for attrib, value in self.attrib.items():
            contour_group_et.set(attrib, str(value))
        
        timestep = et.SubElement(contour_group_et, "timestep")
        timestep.set("id", "0")
        
        lofting_parameters_et = et.SubElement(timestep, 'lofting_parameters')
        for attrib, value in self.lofting_parameters.items():
            lofting_parameters_et.set(attrib, str(value))
        
        for contour in self.contours:
            contour_et = et.SubElement(timestep, 'contour')
            for attrib, value in contour.attrib.items():
                contour_et.set(attrib, str(value))
            
            # path_point
            path_point_et = et.SubElement(contour_et, 'path_point')
            path_point_et.set("id", str(contour.path_point.id))
            
            pos_et = et.SubElement(path_point_et, 'pos')
            pos_et.set("x", str(contour.path_point.pos[0]))
            pos_et.set("y", str(contour.path_point.pos[1]))
            pos_et.set("z", str(contour.path_point.pos[2]))
            
            tangent_et = et.SubElement(path_point_et, 'tangent')
            tangent_et.set("x", str(contour.path_point.tangent[0]))
            tangent_et.set("y", str(contour.path_point.tangent[1]))
            tangent_et.set("z", str(contour.path_point.tangent[2]))
            
            rotation_et = et.SubElement(path_point_et, 'rotation')
            rotation_et.set("x", str(contour.path_point.rotation[0]))
            rotation_et.set("y", str(contour.path_point.rotation[1]))
            rotation_et.set("z", str(contour.path_point.rotation[2]))
            
            # control_points
            control_points_et = et.SubElement(contour_et, 'control_points')
            for i, point in enumerate(contour.control_points):
                point_et = et.SubElement(control_points_et, 'point')
                point_et.set("id", str(i))
                point_et.set("x", str(point[0]))
                point_et.set("y", str(point[1]))
                point_et.set("z", str(point[2]))
                
            # contour_points
            contour_points_et = et.SubElement(contour_et, 'contour_points')
            for i, point in enumerate(contour.contour_points):
                point_et = et.SubElement(contour_points_et, 'point')
                point_et.set("id", str(i))
                point_et.set("x", str(point[0]))
                point_et.set("y", str(point[1]))
                point_et.set("z", str(point[2]))
        
        indent(contour_group_et)
        with open(file_name, "w") as f:
            f.write(  "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" \
                    + "<format version=\"1.0\" />\n" \
                    + et.tostring(contour_group_et).decode("utf-8"))

    def scale(self, contour_ids, scales):
        
        for contour_id, scale in zip(contour_ids, scales):
            contour_id, scale = int(contour_id), float(scale)
            
            #skip iteration, if scale = 1.0
            if scale == 1.0:
                continue
            
            contour = self.contours[contour_id]
            
            # The first control point is the contour center.
            p0 = contour.control_points[0]  #nparray
            for i,control_point in enumerate(contour.control_points):
                contour.control_points[i] = scale*(control_point - p0) + p0
            for i, contour_point in enumerate(contour.contour_points):
                contour.contour_points[i] = scale*(contour_point - p0) + p0


class Contour(object):
    """
    The Contour class is used to represent a SimVascular contour.
    """
    def __init__(self):
        self.attrib = dict()
        
        self.path_point = Path_Point()
        self.control_points = []
        self.contour_points = []
    
    def read(self, contour_et):
        
        self.attrib["id"] = contour_et.attrib.get("id")
        self.attrib["type"] = contour_et.attrib.get("type")
        self.attrib["method"] = contour_et.attrib.get("method")
        self.attrib["closed"] = contour_et.attrib.get("closed")
        self.attrib["min_control_number"] = contour_et.attrib.get("min_control_number")
        self.attrib["max_control_number"] = contour_et.attrib.get("max_control_number")
        self.attrib["subdivision_type"] = contour_et.attrib.get("subdivision_type")
        self.attrib["subdivision_number"] = contour_et.attrib.get("subdivision_number")
        self.attrib["subdivision_spacing"] = contour_et.attrib.get("subdivision_spacing")
        
        path_point_et = contour_et.find("./path_point")
        self.path_point.read(path_point_et)
        
        for control_pts in contour_et.iter('control_points'):
            #logger.info("      Control points ")
            for point in control_pts.iter('point'):
                x = point.attrib['x']
                y = point.attrib['y']
                z = point.attrib['z']
                point = np.array([float(x), float(y), float(z)])
                self.control_points.append(point)
        for contour_pts in contour_et.iter('contour_points'):
            #logger.info("      Contour points ")
            for point in contour_pts.iter('point'):
                x = point.attrib['x']
                y = point.attrib['y']
                z = point.attrib['z']
                point = np.array([float(x), float(y), float(z)])
                self.contour_points.append(point)
            #_for point in control_pts


class Path(object):
    """
    The Path class is used to represent a SimVascular path.
    """
    
    def __init__(self, path_name):
        self.id = path_name
        self.pos = []
        self.tangent = []
        self.rotation = []

class Path_Point(object):
    """
    The Path_Point class is used to represent a SimVascular path_point.
    """
    
    def __init__(self):
        self.id = None
        self.pos = np.zeros(3)
        self.tangent = np.zeros(3)
        self.rotation = np.zeros(3)
    
    def read(self, path_point_et):
        self.id = path_point_et.attrib["id"]
        
        pos_et = path_point_et.find("./pos")
        self.pos[0] = float(pos_et.attrib["x"])
        self.pos[1] = float(pos_et.attrib["y"])
        self.pos[2] = float(pos_et.attrib["z"])
        
        pos_et = path_point_et.find("./tangent")
        self.tangent[0] = float(pos_et.attrib["x"])
        self.tangent[1] = float(pos_et.attrib["y"])
        self.tangent[2] = float(pos_et.attrib["z"])
        
        pos_et = path_point_et.find("./rotation")
        self.rotation[0] = float(pos_et.attrib["x"])
        self.rotation[1] = float(pos_et.attrib["y"])
        self.rotation[2] = float(pos_et.attrib["z"])

def indent(elem, level=0):
  i = "\n" + level*"    "
  if len(elem):
    if not elem.text or not elem.text.strip():
      elem.text = i + "    "
    if not elem.tail or not elem.tail.strip():
      elem.tail = i
    for elem in elem:
      indent(elem, level+1)
    if not elem.tail or not elem.tail.strip():
      elem.tail = i
  else:
    if level and (not elem.tail or not elem.tail.strip()):
      elem.tail = i
 


def parse_args():
    """
    Parse command-line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-s",   "--surface-file",    help="Input surface (.vtp) file.")
    parser.add_argument("-c",   "--contour-file",    help="Input contour (.ctr) file.")
    args = parser.parse_args()

    if len(sys.argv) == 1:
       parser.print_help()
       sys.exit(1)

    return args


def init_logging():
    logger_name = 'sv-contours'
    logger = logging.getLogger(logger_name)
    
    # clear logger, if it is alrerady running.
    if (logger.hasHandlers()):
        logger.handlers.clear()
    
    logger.setLevel(logging.INFO)
    console_handler = logging.StreamHandler()
    formatter = logging.Formatter('[%(name)s] %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    
    return logger 



def read_vtp_file(file_name, logger):
    """
    Read a vtp file. 
    """

    if file_name == None:
        logger.error("No input surface file name given.")
        sys.exit(1)
    else:
        logger.info("Input surface file name: %s" % file_name)

    reader = vtk.vtkXMLPolyDataReader()
    reader.SetFileName(file_name)
    reader.Update()
    poly_data = reader.GetOutput()

    points = poly_data.GetPoints()
    num_points = points.GetNumberOfPoints()
    logger.info("Number of points %d" % num_points)

    polygons = poly_data.GetPolys()
    num_polys = polygons.GetNumberOfCells()
    logger.info("Number of polygons %d" % num_polys)

    return poly_data

def create_contour_geometry(renderer, contour_group, color=(1.0, 1.0, 1.0)):
    """ Create geometry for the contours in a contour group.
    """
    for contour in contour_group.contours:
        contour_points = contour.contour_points
        num_pts = len(contour_points)

        # Create contour geometry points and line connectivity.
        points = vtk.vtkPoints()
        points.SetNumberOfPoints(num_pts)
        lines = vtk.vtkCellArray()
        lines.InsertNextCell(num_pts + 1)
        n = 0
        for pt in contour_points:
            points.SetPoint(n, pt[0], pt[1], pt[2])
            lines.InsertCellPoint(n)
            n += 1
        #_for pt in contour_points
        lines.InsertCellPoint(0)

        geom = vtk.vtkPolyData()
        geom.SetPoints(points)
        geom.SetLines(lines)
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputData(geom)
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(color)
        renderer.AddActor(actor)


def create_graphics_geometry(geom):
    """ Create geometry for display.
    """
    mapper = vtk.vtkPolyDataMapper()
    #mapper.SetInputConnection(geom)
    mapper.SetInputData(geom)
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    #actor.GetProperty().SetLineWidth(3)
    #actor.GetProperty().EdgeVisibilityOn()
    return actor

def main():
    # Get command-line arguments.
    args = parse_args()
    contour_fileName = args.contour_file
    surface_fileName = args.surface_file
    
    # Set logger
    logger = init_logging()

    # Create renderer and graphics window.
    renderer = vtk.vtkRenderer()
    renderer_win = vtk.vtkRenderWindow()
    renderer_win.AddRenderer(renderer)
    renderer.SetBackground(0.5, 0.8, 0.8)
    renderer_win.SetSize(500, 500)

    # Read surface .vtp file and create render geometry.
    surface = read_vtp_file(surface_fileName, logger)
    surface_geom = create_graphics_geometry(surface)
    surface_geom.GetProperty().SetOpacity(0.5)
    renderer.AddActor(surface_geom)

    # Read the contours .ctgr file.
    contour_group = ContourGroup(logger)
    contour_group.read(contour_fileName)
    contour_group2 = copy.deepcopy(contour_group)


#    for i, contour in enumerate(contour_group2.contours):
#        for j, point in enumerate(contour.contour_points):
#            x = point[0]
#            y = point[1]
#            z = point[2]
#            
#            contour_group2.contours[i].contour_points[j] = [2.0*x, 2.0*y, 2.0*z]
    
    contour_group2.scale([1,5,7, 10], [0.5, 1.4, 1.0, 2.0])
    
    contour_group2.write("./out.ctgr")
    
    create_contour_geometry(renderer, contour_group)
    create_contour_geometry(renderer, contour_group2, color=(0.0, 1.0, 0.0))
#    for contour_group in contour_groups2:
#        create_contour_geometry(renderer, contour_group, color=(0.0,1.0,0.0))

    # Create a trackball interacter to transoform the geometry using the mouse.
    interactor = vtk.vtkRenderWindowInteractor()
    interactor.SetInteractorStyle(vtk.vtkInteractorStyleTrackballCamera())
    interactor.SetRenderWindow(renderer_win)
    interactor.Start()
    
    del(logger)



if __name__ == "__main__":
    main()



